from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from clyjin.base.module import Module
    from clyjin.base.plugin import Plugin


class NoModulesPluginError(Exception):
    """
    Plugin contains no modules.
    """
    def __init__(self, PluginClass: type["Plugin"]) -> None:
        super().__init__(
            f"plugin <{PluginClass.get_str()}> defines no modules",
        )


class ForeignModulePluginError(Exception):
    """
    Plugin interacted with module it doesn't have.
    """
    def __init__(
        self,
        PluginClass: type["Plugin"],
        ModuleClass: type["Module"],
    ) -> None:
        super().__init__(
            f"plugin <{PluginClass.get_name()}> does not have module"
            f" <{ModuleClass.get_str()}>",
        )


class DuplicateRootModulePluginError(Exception):
    """
    Plugin has several root modules.
    """
    def __init__(
        self,
        PluginClass: type["Plugin"],
        ModuleClass: type["Module"],
    ) -> None:
        super().__init__(
            f"cannot add root module <{ModuleClass.get_str()}>:"
            f" plugin <{PluginClass.get_str()}> already has a root module",
        )



import asyncio
from pathlib import Path

from clyjin.core.boot import Boot
from clyjin.log import Log


@Log.catch
def main() -> None:
    asyncio.run(Boot(
        rootdir=Path.cwd(),
    ).start())


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path  # noqa: TCH003
from clyjin.base.module import Module  # noqa: TCH003
from typing import TYPE_CHECKING

from clyjin.base.model import Model


class PluginInitializeData(Model):
    root_dir: Path
    config_path: Path
    called_module: Module
    called_plugin_sysdir: Path
    called_plugin_common_sysdir: Path
    called_module_sysdir: Path

    class Config:
        arbitrary_types_allowed = True



from typing import TYPE_CHECKING

from antievil import ExpectedTypeError, NotFoundError, PleaseDefineError

from clyjin.base.errors import (
    DuplicateRootModulePluginError,
    ForeignModulePluginError,
    NoModulesPluginError,
)
from clyjin.base.plugininitializedata import PluginInitializeData

if TYPE_CHECKING:
    from clyjin.base.module import Module


class Plugin:
    """
    Basic importable unit to add new system's functionality.

    Used only as a class.

    Attributes:
        NAME:
            Plugin's name used as top-level namespace to all Plugin's commands.
        MODULE_CLASSES:
            List of Module classes registered by this plugin.
        VERSION(optional):
            Version of the Plugin. Defaults to NONE.

    @abstract
    """
    NAME: str | None = None
    MODULE_CLASSES: list[type["Module"]] | None = None
    VERSION: str | None = None

    _RootModule: type["Module"] | None = None

    def __init__(self) -> None:
        raise NotImplementedError

    @classmethod
    async def initialize(
        cls,
        data: PluginInitializeData,
    ) -> None:
        """
        Method called to initialize a user's plugin.

        @abstract
        """
        return

    @classmethod
    def get_str(cls) -> str:
        return \
            f"Plugin Class <{cls.get_name()}>, version <{cls.get_version()}>"

    @classmethod
    def get_name(cls) -> str:
        if cls.NAME is None:
            raise PleaseDefineError(
                cannot_do=f"plugin <{cls}> initialization",
                please_define="attribute NAME",
            )
        elif not isinstance(cls.NAME, str):
            raise ExpectedTypeError(
                obj=cls.NAME,
                ExpectedType=str,
                is_instance_expected=True,
                ActualType=type(cls.NAME),
            )

        return cls.NAME.strip().lower()

    @classmethod
    def get_module_classes(cls) -> list[type["Module"]]:
        if cls.MODULE_CLASSES is None:
            raise PleaseDefineError(
                cannot_do=f"plugin <{cls}> initialization",
                please_define="attribute MODULE_CLASSES",
            )
        elif not isinstance(cls.MODULE_CLASSES, list):
            raise ExpectedTypeError(
                obj=cls.MODULE_CLASSES,
                ExpectedType=list,
                is_instance_expected=True,
                ActualType=type(cls.MODULE_CLASSES),
            )
        elif len(cls.MODULE_CLASSES) == 0:
            raise NoModulesPluginError(cls)

        return cls.MODULE_CLASSES

    @classmethod
    def get_version(cls) -> str:
        if cls.VERSION is None:
            return "unversioned"
        elif not isinstance(cls.VERSION, str):
            raise ExpectedTypeError(
                obj=cls.VERSION,
                ExpectedType=str,
                is_instance_expected=True,
                ActualType=type(cls.VERSION),
            )

        return cls.VERSION

    @classmethod
    def get_namespaced_module_name(cls, ModuleClass: type["Module"]) -> str:
        """
        Returns Module's name prefixed by Plugin's name.

        If Module's name is `_root`, the Module is considered to be root of
        it's parent Plugin and thus be available under Plugin's name.

        Each Plugin can contain only one root Module.
        """
        cls._check_has_module(ModuleClass)
        return cls._get_namespaced_module_name_nocheck(ModuleClass)

    @classmethod
    def get_module_class(cls, name: str) -> type["Module"]:
        # TODO(ryzhovalex): cache search results for better performance
        # 0

        for ModuleClass in cls.get_module_classes():
            if ModuleClass.cls_get_name() == name:
                return ModuleClass

        raise NotFoundError(
            title="module class with namespaced name",
            value=name,
            options={
                "plugin_class": cls.get_str(),
            },
        )

    @classmethod
    def _check_has_module(cls, ModuleClass: type["Module"]) -> None:
        if ModuleClass not in cls.get_module_classes():
            raise ForeignModulePluginError(
                PluginClass=cls,
                ModuleClass=ModuleClass,
            )

    @classmethod
    def _get_namespaced_module_name_nocheck(
        cls,
        ModuleClass: type["Module"],
    ) -> str:
        module_name: str = ModuleClass.cls_get_name()
        if module_name == "_root":
            cls._set_root_module_class(ModuleClass)
            return cls.get_name()
        return cls.get_name() + "." + module_name

    @classmethod
    def _set_root_module_class(cls, ModuleClass: type["Module"]) -> None:
        if cls._RootModule is not None and cls._RootModule is not ModuleClass:
            raise DuplicateRootModulePluginError(
                PluginClass=cls,
                ModuleClass=ModuleClass,
            )
        cls._RootModule = ModuleClass



import re


def snakefy(name: str) -> str:
    """Converts name to snake_case.

    Note that this is not reversible using camelfy().

    Args:
        name:
            Name to convert.

    Returns:
        Name converted.
    """
    # Reference: https://stackoverflow.com/a/1176023/14748231
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("__([A-Z])", r"_\1", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()



from pathlib import Path

from clyjin.base.model import Model
from clyjin.base.module import Module
from clyjin.base.moduleargs import ModuleArgs
from clyjin.base.plugin import Plugin


class CLIArgs(Model):
    """
    Args parsed by first core CLI layer.

    Attributes:
        ModuleClass:
            Target Module Class.
        PluginClass:
            Target Plugin Class.
        populated_module_args:
            Parsed object of args directed to the target module.
        config_path(optional):
            Path to main configuration file. Defaults to None, i.e. hasn't been
            set.
        verbosity_level(optional):
            How verbose printed output should be. Defaults to 0.
    """
    ModuleClass: type[Module]
    PluginClass: type[Plugin]
    populated_module_args: ModuleArgs | None
    config_path: Path | None
    verbosity_level: int
    sysdir: Path | None

    class Config:
        arbitrary_types_allowed = True



import argparse
import typing
from typing import Any

from antievil import (
    ExpectedTypeError,
    LogicError,
    NotFoundError,
    UnsupportedError,
)

from clyjin.base.module import Module
from clyjin.base.moduleargs import ModuleArg, ModuleArgs
from clyjin.base.plugin import Plugin
from clyjin.core.cli.cliargs import CLIArgs
from clyjin.core.cli.generator import CLIGenerator

if typing.TYPE_CHECKING:
    from pathlib import Path


class CLIParser:
    """
    Parser CLI args into application objects.

    Attributes:
        RegisteredPlugins:
            Plugin classes registered in the system to initialize parser
            groups for.
    """
    def __init__(self, RegisteredPlugins: list[type[Plugin]]) -> None:
        self._RegisteredPlugins: list[type[Plugin]] = RegisteredPlugins
        self._parser: argparse.ArgumentParser = CLIGenerator().get_parser(
            self._RegisteredPlugins,
        )

    def parse(
        self,
        args: list[str] | None = None,
    ) -> CLIArgs:
        """
        Parses incoming CLI args either from `args` argument or from system's
        args.

        Args:
            args(optional):
                List of string args to parse. System args are parsed by
                default.
        """
        namespace: argparse.Namespace = self._parser.parse_args(args)

        PluginClass: type[Plugin]
        ModuleClass: type[Module]
        PluginClass, ModuleClass = \
            self._get_plugin_module_classses_from_input_name(
                namespace.module,
            )

        populated_module_args: ModuleArgs | None = \
            self._populate_module_args_from_namespace(
                ModuleClass,
                namespace,
            )
        config_path: Path | None = namespace.config_path
        verbosity_level: int = namespace.verbosity_level
        sysdir: Path | None = namespace.sysdir

        return CLIArgs(
            ModuleClass=ModuleClass,
            PluginClass=PluginClass,
            populated_module_args=populated_module_args,
            config_path=config_path,
            verbosity_level=verbosity_level,
            sysdir=sysdir,
        )

    def _get_plugin_module_classses_from_input_name(
        self,
        input_name: str,
    ) -> tuple[type[Plugin], type[Module]]:
        PluginClass: type[Plugin]
        ModuleClass: type[Module]

        plugin_name: str
        module_name: str
        input_name_dots_count: int = input_name.count(".")
        if input_name_dots_count == 0:
            plugin_name = input_name
            module_name = "_root"
        elif input_name_dots_count == 1:
            plugin_name, module_name = input_name.split(".")
        else:
            raise UnsupportedError(
                title="more than one separation dot in input module name",
                value=input_name,
            )

        for PC in self._RegisteredPlugins:
            if PC.get_name() == plugin_name:
                PluginClass = PC
                ModuleClass = PluginClass.get_module_class(module_name)
                return PluginClass, ModuleClass

        raise NotFoundError(
            title="registered plugin for namespaced name",
            value=input_name,
        )

    def _populate_module_args_from_namespace(
        self,
        ModuleClass: type[Module],
        namespace: argparse.Namespace,
    ) -> ModuleArgs | None:
        empty_module_args: ModuleArgs | None = ModuleClass.ARGS

        if empty_module_args is None:
            # nothing to populate, left as it is
            return empty_module_args

        empty_module_args = typing.cast(ModuleArgs, empty_module_args)
        populated_module_args: ModuleArgs = empty_module_args.model_copy(
            deep=True,
        )

        for arg_name, _module_arg in empty_module_args.model_dump().items():
            if not isinstance(arg_name, str):
                raise ExpectedTypeError(
                    obj=arg_name,
                    ExpectedType=str,
                    is_instance_expected=True,
                    ActualType=type(arg_name),
                )

            module_arg: ModuleArg = ModuleArg.parse_obj(_module_arg)

            try:
                arg_value: Any = vars(namespace)[arg_name]
            except KeyError as error:
                error_message: str = \
                    f"cannot find argument with name <{arg_name}>" \
                    " in generated namespace"
                raise LogicError(error_message) from error

            if (
                # arg value set to None if it is not defined, which should be
                # handled at the called module
                arg_value is not None
                and not isinstance(arg_value, module_arg.type)
            ):
                raise ExpectedTypeError(
                    obj=arg_value,
                    ExpectedType=module_arg.type,
                    is_instance_expected=True,
                    ActualType=type(arg_value),
                )

            try:
                # finally set value for the module args object, i.e. populate
                getattr(populated_module_args, arg_name).value = arg_value
            except ValueError as error:  # arg_name not found
                error_message: str = \
                    f"cannot find argument with name <{arg_name}>" \
                    f" in model args <{empty_module_args}>"
                raise LogicError(error_message) from error

        return populated_module_args



from pathlib import Path
from typing import Generic

from pydantic.generics import GenericModel

# we should import Config and ModuleArgs since otherwise Pydantic will give
# an undefined error for our generic usage
from clyjin.base.config import Config, ConfigType  # noqa: F401
from clyjin.base.moduleargs import ModuleArgs, ModuleArgsType  # noqa: F401
from clyjin.base.plugin import Plugin


class ModuleData(GenericModel, Generic[ModuleArgsType, ConfigType]):
    """
    Attributes:
        name:
            Module's name either taken from the `NAME` attribute or by
            snakefying the class's name replacing `Module` suffix.
        ParentPlugin:
            Parent Plugin of this Module.
        description:
            Module's description.
        args:
            Module's parsed args with actual values attached
            if the `ARGS` attribute is not set.
        config:
            Parsed instance of class defined in `CONFIG_CLASS` attribute.
        rootdir:
            From where the module was called from.
        plugin_common_sysdir:
            System directory for Plugin's common files used by all Plugin's
            modules.
        module_sysdir:
            System directory for module's files. Your module should use only
            this directory for long-term file-storage needs.
        verbosity_level:
            How verbose module's messages should be.
    """
    name: str
    ParentPlugin: type[Plugin]
    description: str | None
    args: ModuleArgsType | None
    config: ConfigType | None
    plugin_common_sysdir: Path
    module_sysdir: Path
    rootdir: Path
    verbosity_level: int



from clyjin.base.config import Config
from clyjin.base.module import Module
from clyjin.core.plugin.args import ConfiguratorCoreArgs
from clyjin.log import Log


class ConfiguratorModule(Module[ConfiguratorCoreArgs, Config]):
    NAME = "configure"
    DESCRIPTION = "configure the system's core"
    ARGS = ConfiguratorCoreArgs(
    )

    async def execute(self) -> None:
        Log.info("[core.configurator] Hello!")



import importlib.util
import os
import pkgutil
from pathlib import Path
from typing import TYPE_CHECKING, Any

from antievil import ExpectedTypeError, NotFoundError

from clyjin.base.moduledata import ModuleData
from clyjin.base.plugin import Plugin
from clyjin.base.plugininitializedata import PluginInitializeData
from clyjin.core.cli.cliargs import CLIArgs
from clyjin.core.cli.parser import CLIParser
from clyjin.core.plugin.plugin import CorePlugin
from clyjin.log import Log

if TYPE_CHECKING:
    from types import ModuleType as PyModuleType

    from clyjin.base.module import Module


class Boot:
    """
    Central entry unit of application execution.
    """
    def __init__(self, *, rootdir: Path) -> None:
        self._RegisteredPlugins: list[type[Plugin]] = []
        self._config_path: Path
        self._sysdir: Path
        self._called_plugin_sysdir: Path
        self._called_plugin_common_sysdir: Path
        self._called_module_sysdir: Path

        self._root_dir: Path = rootdir

        self._DEFAULT_SYSDIR: Path = Path(
            os.environ["HOME"],
            ".clyjin",
        )
        self._DEFAULT_CONFIG_PATH: Path = Path(
            self._root_dir,
            "clyjin.yml",
        )

    async def start(self) -> None:
        await self._collect_registered_plugins()
        cli_args: CLIArgs = CLIParser(
            self._RegisteredPlugins,
        ).parse()
        self._initialize_paths(cli_args)

        module: Module = cli_args.ModuleClass(ModuleData(
            name=cli_args.ModuleClass.cls_get_name(),
            ParentPlugin=cli_args.PluginClass,
            description=cli_args.ModuleClass.DESCRIPTION,
            args=cli_args.populated_module_args,
            # TODO(ryzhovalex): implement configs
            # 0
            config=None,
            rootdir=self._root_dir,
            plugin_common_sysdir=self._called_plugin_common_sysdir,
            module_sysdir=self._called_module_sysdir,
            verbosity_level=cli_args.verbosity_level,
        ))

        Log.info(
            f"[core] initializing plugin <{cli_args.PluginClass.get_str()}>",
        )
        await cli_args.PluginClass.initialize(PluginInitializeData(
            root_dir=self._root_dir,
            config_path=self._config_path,
            called_module=module,
            called_plugin_sysdir=self._called_plugin_sysdir,
            called_plugin_common_sysdir=self._called_plugin_common_sysdir,
            called_module_sysdir=self._called_module_sysdir,
        ))
        Log.info(
            f"[core] initialized plugin <{cli_args.PluginClass.get_str()}>",
        )

        Log.info(
            f"[core] executing module <{module}>",
        )
        await module.execute()
        Log.info(
            f"[core] executed module <{module}>",
        )

    def _initialize_paths(self, cli_args: CLIArgs) -> None:
        self._sysdir = \
            self._DEFAULT_SYSDIR \
            if cli_args.sysdir is None else cli_args.sysdir

        self._called_plugin_sysdir = Path(
            self._sysdir,
            "plugins",
            cli_args.PluginClass.get_name(),
        )

        self._called_plugin_common_sysdir = Path(
            self._called_plugin_sysdir,
            "common",
        )

        self._called_module_sysdir = Path(
            self._called_plugin_sysdir,
            cli_args.ModuleClass.cls_get_name(),
        )

        self._sysdir.mkdir(parents=True, exist_ok=True)
        self._called_plugin_sysdir.mkdir(parents=True, exist_ok=True)
        self._called_plugin_common_sysdir.mkdir(parents=True, exist_ok=True)
        self._called_module_sysdir.mkdir(parents=True, exist_ok=True)

        self._config_path = \
            self._DEFAULT_CONFIG_PATH \
            if cli_args.config_path is None else cli_args.config_path
        if not self._config_path.exists():
            Log.warning(
                f"[core] config is not found at <{self._config_path}>:"
                " use defaults",
            )

    async def _collect_registered_plugins(self) -> None:
        # always add Core Plugin
        self._RegisteredPlugins.append(CorePlugin)
        Log.info(
            f"[core] loaded core plugin <{CorePlugin.get_str()}>",
        )

        for finder, name, _ispkg in pkgutil.iter_modules():
            if name.startswith("clyjin_"):
                pathstr: str = self._get_finder_pathstr(finder)
                Log.info(
                    f"[core] found Python module <{name}> at <{pathstr}>",
                )

                try:
                    LoadedPlugin: type[Plugin] = self._load_plugin(name)
                except (NotFoundError, ExpectedTypeError) as error:
                    Log.error(
                        "[core] failed to load plugin"
                        f" <{name}>: error=<{error}>",
                    )
                    continue

                Log.info(
                    f"[core] loaded plugin <{LoadedPlugin.get_str()}>",
                )

    def _get_finder_pathstr(self, finder: Any) -> str:
        try:
            return str(finder.path)
        except AttributeError:
            return "unattached path"

    def _load_plugin(self, name: str) -> type[Plugin]:
        imported_module: PyModuleType = importlib.import_module(name)
        try:
            # MainPlugin variable is searched by default, maybe later it might
            # be configurable
            ImportedMainPlugin: type[Module] = imported_module.MainPlugin
        except AttributeError as error:
            raise NotFoundError(
                title="MainPlugin attribute of imported module",
                value=imported_module,
            ) from error

        if not issubclass(ImportedMainPlugin, Plugin):
            raise ExpectedTypeError(
                obj=ImportedMainPlugin,
                ExpectedType=Plugin,
                # TODO(ryzhovalex): use `expected_inheritance` to denote
                #   issubclass() usage as it gets support at Antievil
                # 0
                is_instance_expected=True,
                ActualType=type(ImportedMainPlugin),
            )

        self._RegisteredPlugins.append(ImportedMainPlugin)
        return ImportedMainPlugin



import argparse
from argparse import _SubParsersAction as ArgparseSubParsersAction
from pathlib import Path
from typing import Any

from antievil import ExpectedTypeError, UnsupportedError

from clyjin.base.module import Module
from clyjin.base.moduleargs import ModuleArg, ModuleArgs
from clyjin.base.plugin import Plugin


class CLIGenerator:
    """
    Generates argument parser for registered modules.
    """
    def get_parser(
        self,
        RegisteredPlugins: list[type[Plugin]],
    ) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser = argparse.ArgumentParser(
            description="Clyjin",
        )

        self._add_common_args(parser)
        self._module_subparser_hub: ArgparseSubParsersAction = \
            self._add_module_subparser_hub(parser)
        self._add_plugins_parsers(RegisteredPlugins)

        return parser

    def _add_common_args(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="verbosity level",
            dest="verbosity_level",
        )
        parser.add_argument(
            "-c",
            "--config",
            type=Path,
            default=None,
            help=
                "path to config file. Defaults to `clyjin.yml` in current dir",
            dest="config_path",
        )
        parser.add_argument(
            "--sysdir",
            type=Path,
            default=None,
            help=
                "directory for clyjin global configs."
                " Defaults to `$HOME/.clyjin`",
            dest="sysdir",
        )

    def _add_module_subparser_hub(
        self,
        parser: argparse.ArgumentParser,
    ) -> ArgparseSubParsersAction:
        return parser.add_subparsers(
            help="module to launch",
            dest="module",
        )

    def _add_plugins_parsers(
        self,
        RegisteredPlugins: list[type[Plugin]],
    ) -> None:
        for PluginClass in RegisteredPlugins:
            self._add_plugin_modules_parsers(
                PluginClass,
            )

    def _add_plugin_modules_parsers(
        self,
        PluginClass: type[Plugin],
    ) -> None:
        for ModuleClass in PluginClass.get_module_classes():
            self._add_plugin_module_parser(PluginClass, ModuleClass)

    def _add_plugin_module_parser(
        self,
        PluginClass: type[Plugin],
        ModuleClass: type[Module],
    ) -> None:
        # add main parser in any case
        module_parser: argparse.ArgumentParser = \
            self._module_subparser_hub.add_parser(
                # prefix any module name with Plugin's namespace
                PluginClass.get_namespaced_module_name(ModuleClass),
                help=ModuleClass.DESCRIPTION,
            )

        module_args: ModuleArgs | None = ModuleClass.ARGS
        if module_args is None:
            # register modules without args only with initial keyword
            return
        self._parse_module_args(module_args, module_parser)

    def _parse_module_args(
        self,
        module_args: ModuleArgs,
        module_parser: argparse.ArgumentParser,
    ) -> None:
        for arg_name, _module_arg in module_args.model_dump().items():
            if not isinstance(arg_name, str):
                raise ExpectedTypeError(
                    obj=arg_name,
                    ExpectedType=str,
                    is_instance_expected=True,
                    ActualType=type(arg_name),
                )

            module_arg: ModuleArg = ModuleArg.parse_obj(_module_arg)

            argparse_type: type | None  = \
                module_arg.type \
                if module_arg.argparse_type is None \
                else module_arg.argparse_type

            arg_add_optionals: dict[str, Any] = dict(
                type=argparse_type,
                action=module_arg.action,
                nargs=module_arg.nargs,
                const=module_arg.const,
                default=module_arg.default,
                choices=module_arg.choices,
                required=module_arg.required,
                help=module_arg.help,
                metavar=module_arg.metavar,
                **module_arg.argparse_kwargs
                    if module_arg.argparse_kwargs else {},
            )

            if argparse_type is type:
                del arg_add_optionals["type"]

            # do not supply `dest` for positional arguments - argparse gives
            # an error for that
            if module_arg.is_optional():
                module_parser.add_argument(
                    *module_arg.names,
                    dest=arg_name,
                    **arg_add_optionals,
                )
            else:
                # positional arguments are always required
                if (
                    arg_add_optionals["required"] is not None
                    and arg_add_optionals["required"] is False
                ):
                    raise UnsupportedError(
                        title="non-required positional with name",
                        value=arg_name,
                    )
                del arg_add_optionals["required"]

                module_parser.add_argument(
                    *module_arg.names,
                    **arg_add_optionals,
                )



from typing import TYPE_CHECKING, Generic, TypeVar

from antievil import (
    CannotBeNoneError,
    ExpectedTypeError,
    PleaseDefineError,
)

from clyjin.base.config import ConfigType
from clyjin.base.moduleargs import ModuleArgsType

if TYPE_CHECKING:
    from pathlib import Path

    from clyjin.base.moduledata import ModuleData
    from clyjin.base.plugin import Plugin

ModuleType = TypeVar("ModuleType", bound="Module")
class Module(Generic[ModuleArgsType, ConfigType]):
    """
    Basic Clyjin unit for defining own interfaces.

    Every defined module should inherit this class and implement method
    `execute()`.

    After creating own subclass of Module, you can call
    `clyjin core register <path/to/your/module.py>:MyModule` which will
    register a module MyModule in the clyjin's storage.

    A custom-defined Config can be attached to every Module, which is parsed by
    Clyjin from the user's configuration file, AKA `clyjin.yml`.

    In the configuration file, a field with the name, corresponding to the
    Module's name is used to store any configuration data. For example, for a
    Module named `DonutShopModule`, the according field will be named
    `DonutShop`.

    To attach a config, class-attribute `Model.CONFIG` is used. For further
    details of Config making, see according documentation at
    [Config]($(ref.orwynn.core.config.Config)).

    Class-Attributes:
        NAME:
            Name of the Module primarily used as CLI module's name. If
            `_root` is used as a Module's name, the Module is considered to
            be root to it's parent Plugin.
        DESCRIPTION(optional):
            Description of Module primarily appeared in CLI help section.
            Defaults to None.
        ARGS(optional):
            Module Args class attached to the Module. No args are accepted
            by default. Note that these args are not populated with values,
            for populated args see `self.args` property.
        CONFIG_CLASS(optional):
            Config class attached to the Module. No config is attached by
            default.

    Attributes:
        module_data:
            Required data to initialize the Module. Passed by Clyjin System.

    @abstract
    """
    NAME: str | None = None
    DESCRIPTION: str | None = None
    ARGS: ModuleArgsType | None = None
    CONFIG_CLASS: type[ConfigType] | None = None

    def __init__(
        self,
        module_data: "ModuleData[ModuleArgsType, ConfigType]",
    ) -> None:
        self._name: str = module_data.name
        self._description: str | None = module_data.description
        self._args: ModuleArgsType | None = module_data.args
        self._config: ConfigType | None = module_data.config
        self._plugin_common_sysdir: Path = module_data.plugin_common_sysdir
        self._module_sysdir: Path = module_data.module_sysdir
        self._rootdir: Path = module_data.rootdir
        self._verbosity_level: int = module_data.verbosity_level
        self._ParentPlugin: type["Plugin"] = module_data.ParentPlugin

    def __str__(self) -> str:
        return \
            f"Module <{self.cls_get_name()}>" \
            f" of Plugin <{self._ParentPlugin.get_str()}>"

    @classmethod
    def cls_get_str(cls) -> str:
        return f"Module Class <{cls.cls_get_name()}>"

    @classmethod
    def cls_get_name(cls) -> str:
        if cls.NAME is None:
            raise PleaseDefineError(
                cannot_do=f"module <{cls}> initialization",
                please_define="attribute NAME",
            )
        elif not isinstance(cls.NAME, str):
            raise ExpectedTypeError(
                obj=cls.NAME,
                ExpectedType=str,
                is_instance_expected=True,
                ActualType=type(cls.NAME),
            )

        return cls.NAME.strip().lower()

    @property
    def args(self) -> ModuleArgsType:
        if self._args is None:
            raise CannotBeNoneError(
                title=f"in order to retrieve, Module <{self}> args",
            )
        return self._args

    async def execute(self) -> None:
        raise NotImplementedError



from clyjin.base.config import Config
from clyjin.base.model import Model
from clyjin.base.module import Module
from clyjin.base.moduleargs import ModuleArg, ModuleArgs
from clyjin.base.moduledata import ModuleData
from clyjin.base.plugin import Plugin
from clyjin.base.plugininitializedata import PluginInitializeData



from typing import Any, Generic, TypeVar

from antievil import CannotBeNoneError, UnsetValueError
from pydantic.generics import GenericModel

from clyjin.base.model import Model
from clyjin.utils.types import T

ModuleArgsType = TypeVar("ModuleArgsType", bound="ModuleArgs")
class ModuleArgs(Model):
    """
    Args object parsed by CLI and passed to Module.

    While the keys might be freely chosen (but should be strings), the values
    should be of `$(ref.clyjin.base.moduleargs.ModuleArg)` type.

    @abstract
    """


class ModuleArg(GenericModel, Generic[T]):
    """
    Description of Module's Argument, in form similar (except without e.g.
    `dest` argument) to `argparse.ArgumentParser.add_argument()` call.

    This object is also passed back to the host Module after the parsing, with
    according values attached.

    Use `argparse_type` if you want to override `type` as an argument to
    `argparse.add_argument`. It is useful in situations where you want to
    accept list, but don't want to invoke argparse's special logic for list
    type. Pass `argparse_type=type` to disable passing type to argparse at all.
    """
    action: str | None = None
    nargs: int | str | None = None
    const: T | None = None
    default: T | None = None
    choices: list[T] | None = None
    required: bool | None = None
    help: str | None = None
    metavar: str | tuple[str, ...] | None = None
    argparse_kwargs: dict[str, Any] | None = None

    argparse_type: type | None = None

    names: list[str]
    type: type[T]

    _value: T | None = None

    @property
    def value(self) -> T:
        if self._value is None:
            raise UnsetValueError(
                explanation=f"cannot get module arg <{self}> value",
            )
        return self._value

    @value.setter
    def value(self, value: T) -> None:
        if value is None and not self.is_optional():
            raise CannotBeNoneError(
                title=f"on setting module arg <{self}> value",
            )
        self._value = value

    def is_optional(self) -> bool:
        return not any("-" in name for name in self.names)



import clyjin
from clyjin._project import get_version
from clyjin.base.plugin import Plugin
from clyjin.core.plugin.configurator import ConfiguratorModule


class CorePlugin(Plugin):
    NAME = "core"
    MODULE_CLASSES = [
        ConfiguratorModule,
    ]
    VERSION = get_version()

    @classmethod
    def get_version(cls) -> str | None:
        cls.VERSION = clyjin.__version__
        return super().get_version()
